# Copyright (c) YugaByte, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.  See the License for the specific language governing permissions and limitations
# under the License.

import os


def mkdir_p(d):
    """
    Similar to the "mkdir -p ..." shell command. Creates the given directory and all enclosing
    directories. No-op if the directory already exists.
    """

    if os.path.isdir(d):
        return
    try:
        os.makedirs(d)
    except OSError:
        if os.path.isdir(d):
            return
        raise




# Copyright (c) YugaByte, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.  See the License for the specific language governing permissions and limitations
# under the License.


import os
import re


def quote_for_bash(s):
    if s == '':
        return "''"
    if re.search(r"""['"${}()\\]""", s):
        return "'" + s.replace("'", r"'\''") + "'"
    return s


def safe_path_join(*args):
    """
    Like os.path.join, but allows arguments to be None. If all arguments are None, returns None.
    A special case: if the first argument is None, always return None. That allows to set a number
    of constants as relative paths under a certain path which may itself be None.

    >>> safe_path_join()
    >>> safe_path_join(None)
    >>> safe_path_join(None, None)
    >>> safe_path_join('/a', None, 'b')
    '/a/b'
    >>> safe_path_join(None, '/a', None, 'b')  # special case: first arg is None
    """
    if not args or args[0] is None:
        return None
    args = [arg for arg in args if arg is not None]
    return os.path.join(*args)


def cmd_line_args_to_str(args):
    """
    Converts a list of command-line arguments, including an executable program in the beginning,
    to a single command-line string suitable for pasing into Bash.

    :param args: an array with a program path and command-line arguments
    :return: a string suitable for pasing into Bash
    """
    return ' '.join([quote_for_bash(arg) for arg in args])


def trim_long_text(text, max_lines):
    """
    Trim a potentially long multi-line message at the given number of lines.
    :param text: the input text
    :param max_lines: maximum number of lines
    :return: the trimmed message
    """
    max_lines = max(max_lines, 3)

    lines = text.splitlines()
    if len(lines) <= max_lines:
        return text

    # Here is the math involved:
    # lines_at_each_end * 2 + 1 <= max_lines
    # lines_at_each_end <= (max_lines - 1) / 2
    lines_to_keep_at_each_end = int((max_lines - 1) / 2)

    num_lines_skipped = len(lines) - lines_to_keep_at_each_end * 2
    if num_lines_skipped <= 0:
        return text

    return "\n".join(
        lines[:lines_to_keep_at_each_end] +
        ['({} lines skipped)'.format(num_lines_skipped)] +
        lines[-lines_to_keep_at_each_end:]
    )


def decode_utf8(bytes):
    if isinstance(bytes, str):
        return bytes
    return bytes.decode('utf-8')


def get_bool_env_var(env_var_name):
    """
    >>> os.environ['YB_TEST_VAR'] = '  1 '
    >>> get_bool_env_var('YB_TEST_VAR')
    True
    >>> os.environ['YB_TEST_VAR'] = '  0 '
    >>> get_bool_env_var('YB_TEST_VAR')
    False
    >>> os.environ['YB_TEST_VAR'] = '  TrUe'
    >>> get_bool_env_var('YB_TEST_VAR')
    True
    >>> os.environ['YB_TEST_VAR'] = 'fAlSe '
    >>> get_bool_env_var('YB_TEST_VAR')
    False
    >>> os.environ['YB_TEST_VAR'] = '  YeS '
    >>> get_bool_env_var('YB_TEST_VAR')
    True
    >>> os.environ['YB_TEST_VAR'] = 'No'
    >>> get_bool_env_var('YB_TEST_VAR')
    False
    >>> os.environ['YB_TEST_VAR'] = ''
    >>> get_bool_env_var('YB_TEST_VAR')
    False
    """
    value = os.environ.get(env_var_name, None)
    if value is None:
        return False

    return value.strip().lower() in ['1', 't', 'true', 'y', 'yes']




# Copyright (c) YugaByte, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.  See the License for the specific language governing permissions and limitations
# under the License.

"""Utilities for running external commands"""

import os
import logging
import subprocess

from yugabyte_pycommon.text_manipulation import cmd_line_args_to_str, decode_utf8, trim_long_text
from yugabyte_pycommon.logging_util import is_verbose_mode


class ProgramResult:
    """
    This represents the result of executing an external program.
    """
    def __init__(self, cmd_line, returncode, stdout, stderr, error_msg, program_path):
        self.cmd_line = cmd_line
        self.returncode = returncode
        self.stdout = stdout
        self.stderr = stderr
        self.error_msg = error_msg
        self.program_path = program_path

    def success(self):
        return self.returncode != 0


class ExternalProgramError(Exception):
    def __init__(self, message, result):
        self.message = message
        self.result = result


work_dir_context_stack = []


class WorkDirContext:
    """
    Allows setting a context for running external programs. Does not actually change the working
    directory.

    Example:

    with WorkDirContext('/tmp'):
        run_program('ls')
    """
    def __init__(self, work_dir):
        self.work_dir = work_dir

    def __enter__(self):
        work_dir_context_stack.append(self)

    def __exit__(self, exception_type, exception_value, traceback):
        assert self is work_dir_context_stack[-1]
        work_dir_context_stack.pop()


def run_program(args, error_ok=False, max_error_lines=10000, cwd=None, shell=None, **kwargs):
    """
    Run the given program identified by its argument list, and return a ProgramResult object.
    :param error_ok: False to raise an exception on errors, True not to raise it.
    """
    if isinstance(args, tuple):
        args = list(args)

    if isinstance(args, str):
        args = [args]
        if shell is None:
            shell = True

    if cwd is None and work_dir_context_stack:
        cwd = work_dir_context_stack[-1].work_dir

    def normalize_arg(arg):
        if isinstance(arg, int):
            return str(arg)
        return arg

    args = [normalize_arg(arg) for arg in args]

    cmd_line_str = cmd_line_args_to_str(args)
    invocation_details_str = "external program in directory %s: %s" % (
        cwd or os.getcwd(), cmd_line_str)

    if is_verbose_mode():
        logging.info("Running %s", invocation_details_str)

    try:
        program_subprocess = subprocess.Popen(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=shell,
            cwd=cwd,
            **kwargs)

    except OSError:
        logging.error("Failed to run %s", invocation_details_str)
        raise

    program_stdout, program_stderr = program_subprocess.communicate()
    error_msg = None

    def cleanup_output(out_str):
        return decode_utf8(out_str.rstrip())

    clean_stdout = cleanup_output(program_stdout)
    clean_stderr = cleanup_output(program_stderr)

    result = ProgramResult(
        cmd_line=args,
        program_path=os.path.realpath(args[0]),
        returncode=program_subprocess.returncode,
        stdout=clean_stdout,
        stderr=clean_stderr,
        error_msg=error_msg)

    if program_subprocess.returncode != 0:
        error_msg = "Non-zero exit code {} from: {} ; stdout: '{}' stderr: '{}'".format(
                program_subprocess.returncode, cmd_line_str,
                trim_long_text(clean_stdout, max_error_lines),
                trim_long_text(clean_stderr, max_error_lines))
        if not error_ok:
            raise ExternalProgramError(error_msg, result)

    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of yugabyte_pycommon.
# https://github.com/yugabyte/yugabyte_pycommon

# Licensed under the Apache 2.0 license:
# http://www.opensource.org/licenses/Apache 2.0-license
#  Copyright (c) YugaByte, Inc.

from setuptools import setup, find_packages
from yugabyte_pycommon import __version__

tests_require = [
    'mock',
    'nose',
    'coverage',
    'yanc',
    'preggy',
    'tox',
    'ipdb',
    'coveralls',
    'sphinx',
]

setup(
    name='yugabyte_pycommon',
    version=__version__,
    description='Common utilities used in YugaByte Database\'s build infrastructure but could '
                'also be useful for anyone. E.g. convenient utilities for running external '
                'programs, logging, etc. Please give YugaByte DB a star at '
                'https://github.com/yugabyte/yugabyte-db -- much appreciated!',
    long_description=
        'Common utilities used in YugaByte Database\'s build infrastructure but could also be '
        'useful for anyone. E.g. convenient utilities for running external programs, logging, etc.',
    keywords='tool tools utility utilities yugabyte run command external process group_by',
    author='Mikhail Bautin',
    author_email='mbautin@users.noreply.github.com',
    url='https://github.com/yugabyte/yugabyte_pycommon',
    license='Apache 2.0',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: Unix',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Operating System :: OS Independent',
    ],
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        # add your dependencies here
        # remember to use 'package-name>=x.y.z,<x.y+1.0' notation (this way you get bugfixes)
    ],
    extras_require={
        'tests': tests_require,
    },
    entry_points={
        'console_scripts': [
            # add cli scripts here in this form:
            # 'yugabyte_pycommon=yugabyte_pycommon.cli:main',
        ],
    },
)



# Copyright (c) YugaByte, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.  See the License for the specific language governing permissions and limitations
# under the License.
#

from .base import TestCase

import os

from yugabyte_pycommon import run_program, quote_for_bash, ExternalProgramError, WorkDirContext


class ExternalCommandsTestCase(TestCase):
    def test_run_program_noerror(self):
        result = run_program("true")
        self.assertEquals(0, result.returncode)
        self.assertEquals('', result.stdout)
        self.assertEquals('', result.stderr)

    def test_exit_codes(self):
        for exit_code in [0, 1, 2, 3, 10, 20, 100, 150, 200, 250, 255]:
            result = run_program("exit %d" % exit_code, shell=True, error_ok=exit_code != 0)
            self.assertEquals(exit_code, result.returncode)

    def test_quote_for_bash(self):
        for s in [
            '',
            'a',
            'a b'
            '"foo"'
            'foo"bar',
            'foo " bar',
            "foo'bar",
            '$foo $bar',
            "\\",
            "\\\\",
            "\\\\\\",
            '"' + "'",
            '"' + "'" + '"' + "'",
            r"""'\''"""
        ]:
            result = run_program('echo ' + quote_for_bash(s), shell=True)
            self.assertEquals(s, result.stdout)

    def test_error_reporing(self):
        with self.assertRaises(ExternalProgramError):
            run_program('false')

    def test_work_dir_context(self):
        old_work_dir = os.getcwd()
        for d in ['/tmp', os.path.expanduser('~')]:
            with WorkDirContext(d):
                self.assertEquals(d, run_program('pwd').stdout)
                # The current directory should not actually change.
                self.assertEquals(old_work_dir, os.getcwd())
